.xr-wrap {
    font-size: 0.85em;
    margin-left: 1.25em;
    padding-left: 1.25em;
    border-left: thin var(--color-foreground-muted) solid;
  }
  .xr-array-wrap, .xr-var-data, .xr-var-preview {
    font-size: 0.9em;
  }
  .table-wrapper{
    font-family: var(--font-stack--monospace);
    font-size: var(--font-size--small--2);
  }
  .gp {
    color: darkorange;
  }
